# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for eagle_strategy."""

import numpy as np
from vizier import algorithms as vza
from vizier import pyvizier as vz
from vizier._src.algorithms.designers.eagle_strategy import eagle_strategy
from vizier._src.algorithms.designers.eagle_strategy import eagle_strategy_utils
from vizier._src.algorithms.designers.eagle_strategy import testing

from absl.testing import absltest
from absl.testing import parameterized


EagleStrategyDesigner = eagle_strategy.EagleStrategyDesigner
FireflyAlgorithmConfig = eagle_strategy_utils.FireflyAlgorithmConfig


class EagleStrategyTest(parameterized.TestCase):

  @parameterized.parameters(
      testing.create_fake_populated_eagle_designer,  # suggest by mutation.
      testing.create_fake_empty_eagle_designer,  # suggest by initial designer.
  )
  def test_dump_and_load(self, designer_factory):
    eagle_designer = designer_factory()
    metadata = eagle_designer.dump()
    # Create a new eagle designer and load state
    eagle_designer_restored = testing.create_fake_empty_eagle_designer()
    eagle_designer_restored.load(metadata)
    # Generate suggestions from the two designers and test if they're equal.
    for _ in range(10):
      self.assertEqual(
          eagle_designer.suggest()[0].parameters,
          eagle_designer_restored.suggest()[0].parameters,
      )

  def test_load_with_no_state(self):
    problem = testing.create_fake_problem_statement()
    eagle_designer = EagleStrategyDesigner(problem)
    metadata = vz.Metadata()
    # Check that the designer can accept empty metadata in load.
    eagle_designer.load(metadata)

  def test_suggest_one(self):
    eagle_designer = testing.create_fake_populated_eagle_designer()
    trial_suggestion = eagle_designer._suggest_one()
    self.assertIsInstance(trial_suggestion, vz.TrialSuggestion)
    self.assertIsNotNone(
        trial_suggestion.metadata.ns('eagle').get('parent_fly_id')
    )

  def test_embedding(self):
    eagle_designer = testing.create_fake_populated_eagle_designer()
    # Check that the problem was converted.
    self.assertEqual(
        eagle_designer._problem.search_space.parameters[0].bounds, (0.0, 1.0)
    )
    # Check that internal suggestions are in normalized range.
    for _ in range(10):
      trial_suggestion = eagle_designer._suggest_one()
      self.assertBetween(trial_suggestion.parameters['x'].value, 0.0, 1.0)
    # Check that update maps the trials correctly to the normalized space.
    eagle_designer = testing.create_fake_empty_eagle_designer()
    trial = vz.Trial({'x': 10.0})
    trial = trial.complete(
        vz.Measurement(metrics={'objective': np.random.uniform()})
    )
    complete_trials = vza.CompletedTrials([trial])
    eagle_designer.update(complete_trials, vza.ActiveTrials())
    self.assertEqual(
        eagle_designer._firefly_pool._pool[0].trial.parameters['x'].value, 1.0
    )

  @parameterized.parameters(1e-4, 1.0)
  def test_penalize_parent_fly_no_trial_change(self, perturbation):
    eagle_designer = testing.create_fake_populated_eagle_designer(
        x_values=[1.0, 2.0, 3.0], obj_values=[1, 2, 3], parent_fly_ids=[1, 2, 3]
    )
    trial = testing.create_fake_trial(parent_fly_id=2, x_value=2.0, obj_value=2)
    parent_fly = eagle_designer._firefly_pool._pool[2]
    # Set the perturbation.
    parent_fly.perturbation = perturbation
    before_perturbation = parent_fly.perturbation
    eagle_designer._penalize_parent_fly(parent_fly, trial)
    after_perturbation = parent_fly.perturbation
    if perturbation == 1.0:
      # Perturbation is already high so capped by the maximimum.
      self.assertEqual(
          after_perturbation,
          before_perturbation * eagle_designer._config.max_perturbation,
      )
    elif perturbation == 1e-4:
      # Not reaching the maximum yet, multiply by 10.
      self.assertEqual(after_perturbation, 1e-3)

  def test_penalize_parent_fly(self):
    # Capacitated pool size has 11 fireflies.
    eagle_designer = testing.create_fake_populated_eagle_designer(
        x_values=[1.0, 2.0, 3.0], obj_values=[1, 2, 3], parent_fly_ids=[1, 2, 3]
    )
    trial = testing.create_fake_trial(
        parent_fly_id=2, x_value=1.42, obj_value=0.5
    )
    parent_fly = eagle_designer._firefly_pool._pool[2]
    before_perturbation = parent_fly.perturbation
    eagle_designer._penalize_parent_fly(parent_fly, trial)
    after_perturbation = parent_fly.perturbation
    self.assertEqual(after_perturbation, before_perturbation * 0.9)

  def test_suggest(self):
    eagle_designer = testing.create_fake_populated_eagle_designer()
    trial_suggestions = eagle_designer.suggest(count=10)
    self.assertLen(trial_suggestions, 10)
    self.assertIsInstance(trial_suggestions[0], vz.TrialSuggestion)

  def test_suggest_flat(self):
    problem = vz.ProblemStatement()
    root = problem.search_space.root
    root.add_float_param('x1', 0.0, 1.0)
    root.add_float_param('x2', 0.0, 2.0)
    metric = vz.MetricInformation(
        name='obj',
        goal=vz.ObjectiveMetricGoal.MAXIMIZE,
    )
    problem.metric_information.append(metric)

    designer = EagleStrategyDesigner(problem)
    suggestions = designer.suggest(count=1)
    self.assertLen(suggestions, 1)

  def test_update_capacitated_pool_no_parent_fly_trial_is_better(self):
    # Capacitated pool size has 11 fireflies.
    eagle_designer = testing.create_fake_populated_eagle_designer(
        x_values=[1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9, 2.0, 2.1],
        obj_values=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
    )
    trial = testing.create_fake_trial(
        parent_fly_id=98, x_value=1.42, obj_value=100.0
    )
    eagle_designer._update_one(trial)
    self.assertIs(eagle_designer._firefly_pool._pool[3].trial, trial)

  def test_update_capacitated_pool_no_parent_fly_trial_is_not_better(self):
    eagle_designer = testing.create_fake_populated_eagle_designer(
        x_values=[1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9, 2.0, 2.1],
        obj_values=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
    )
    trial = testing.create_fake_trial(
        parent_fly_id=98, x_value=1.42, obj_value=-80.0
    )
    prev_trial = eagle_designer._firefly_pool._pool[3].trial
    eagle_designer._update_one(trial)
    self.assertIs(eagle_designer._firefly_pool._pool[3].trial, prev_trial)

  def test_update_capacitated_pool_with_parent_fly_trial_is_better(self):
    eagle_designer = testing.create_fake_populated_eagle_designer(
        x_values=[1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9, 2.0, 2.1],
        obj_values=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
    )
    trial = testing.create_fake_trial(
        parent_fly_id=2, x_value=3.3, obj_value=80.0
    )
    eagle_designer._update_one(trial)
    self.assertIs(eagle_designer._firefly_pool._pool[2].trial, trial)

  def test_update_capacitated_pool_with_parent_fly_trial_is_not_better(self):
    eagle_designer = testing.create_fake_populated_eagle_designer(
        x_values=[1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9, 2.0, 2.1],
        obj_values=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
    )
    trial = testing.create_fake_trial(
        parent_fly_id=2, x_value=3.3, obj_value=-80.0
    )
    prev_trial = eagle_designer._firefly_pool._pool[2].trial
    eagle_designer._update_one(trial)
    self.assertIs(eagle_designer._firefly_pool._pool[2].trial, prev_trial)

  def test_update_empty_pool(self):
    eagle_designer = testing.create_fake_empty_eagle_designer()
    trial = testing.create_fake_trial(
        parent_fly_id=0, x_value=3.3, obj_value=0.0
    )
    eagle_designer._update_one(trial)
    self.assertIs(eagle_designer._firefly_pool._pool[0].trial, trial)

  def test_seeded_random_samples(self):
    problem = testing.create_fake_problem_statement()
    suggestions1 = EagleStrategyDesigner(problem, seed=1).suggest(count=3)
    suggestions2 = EagleStrategyDesigner(problem, seed=1).suggest(count=3)
    for suggestions1, suggest2 in zip(suggestions1, suggestions2):
      self.assertEqual(suggestions1.parameters, suggest2.parameters)

  @parameterized.parameters(1, 3, 5)
  def test_suggest_update(self, batch_size):
    problem = vz.ProblemStatement()
    problem.search_space.select_root().add_float_param(
        'float1', 1e-2, 1e3, scale_type=vz.ScaleType.LOG
    )
    problem.search_space.select_root().add_float_param(
        'float2', -2.0, 5.0, scale_type=vz.ScaleType.LINEAR
    )
    problem.search_space.select_root().add_int_param(
        'int', min_value=0, max_value=10
    )
    problem.search_space.select_root().add_discrete_param(
        'discrete', feasible_values=[0.0, 0.6]
    )
    problem.search_space.select_root().add_categorical_param(
        'categorical', feasible_values=['a', 'b', 'c']
    )
    problem.metric_information.append(
        vz.MetricInformation(goal=vz.ObjectiveMetricGoal.MINIMIZE, name='')
    )
    eagle_designer = EagleStrategyDesigner(problem)

    tid = 1
    # Simulate running the designer for 3 suggestions each with a batch.
    for _ in range(3):
      suggestions = eagle_designer.suggest(batch_size)
      completed = []
      # Completing the suggestions while assigning unique trial id.
      for suggestion in suggestions:
        completed.append(
            suggestion.to_trial(tid).complete(
                vz.Measurement(metrics={'': np.random.uniform()})
            )
        )
        tid += 1
      eagle_designer.update(vza.CompletedTrials(completed), vza.ActiveTrials())

  @parameterized.named_parameters(
      dict(
          testcase_name='Less suggestions than pool capacity',
          num_feasible_suggestions=3,
          num_infeasible_suggestions=2,
      ),
      dict(
          testcase_name='More suggestions than pool capacity',
          num_feasible_suggestions=50,
          num_infeasible_suggestions=5,
      ),
  )
  def test_infeasible_trials(
      self, num_feasible_suggestions, num_infeasible_suggestions
  ):
    """Tests that Eagle works with infeasible trials."""
    problem = vz.ProblemStatement()
    problem.search_space.select_root().add_float_param(
        'float1', 1e-2, 1e3, scale_type=vz.ScaleType.LOG
    )
    problem.search_space.select_root().add_float_param(
        'float2', -2.0, 5.0, scale_type=vz.ScaleType.LINEAR
    )
    problem.search_space.select_root().add_int_param(
        'int', min_value=0, max_value=10
    )
    problem.search_space.select_root().add_discrete_param(
        'discrete', feasible_values=[0.0, 0.6]
    )
    problem.search_space.select_root().add_categorical_param(
        'categorical', feasible_values=['a', 'b', 'c']
    )
    problem.metric_information.append(
        vz.MetricInformation(goal=vz.ObjectiveMetricGoal.MINIMIZE, name='')
    )
    config = FireflyAlgorithmConfig(infeasible_force_factor=0.1)
    eagle_designer = EagleStrategyDesigner(problem, config=config)

    def _suggest_and_update(
        eagle_designer: EagleStrategyDesigner, tid: int, infeasible: bool
    ):
      suggestion = eagle_designer.suggest(count=1)[0]
      completed = suggestion.to_trial(tid).complete(
          vz.Measurement(metrics={'': np.random.uniform()}),
          infeasibility_reason='infeasible' if infeasible else None,
      )
      eagle_designer.update(
          vza.CompletedTrials([completed]), vza.ActiveTrials()
      )

    # Suggest trials and update designer for less than pool capacity.
    tid = 1
    for _ in range(num_feasible_suggestions):
      _suggest_and_update(eagle_designer, tid, infeasible=False)
      tid += 1

    # Suggest another trial and return it as infeasible.
    for _ in range(num_infeasible_suggestions):
      _suggest_and_update(eagle_designer, tid, infeasible=True)
      tid += 1

    # Test that the pool size is not affected by infeasible trials..
    self.assertEqual(
        eagle_designer._firefly_pool.size,
        min(eagle_designer._firefly_pool.capacity, num_feasible_suggestions),
    )
    # Test that the pool contains infeasible trials.
    self.assertEqual(
        eagle_designer._firefly_pool._infeasible_count,
        num_infeasible_suggestions,
    )
    self.assertEqual(
        sum([
            1
            for firefly in eagle_designer._firefly_pool._pool.values()
            if firefly.trial.infeasible
        ]),
        num_infeasible_suggestions,
    )
    # Suggest more trials while having infeasible trials in the pool.
    for _ in range(3):
      _suggest_and_update(eagle_designer, tid, infeasible=False)
      tid += 1


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Testing utils for Eagle designer."""
from typing import Optional

import numpy as np
from vizier import pyvizier as vz
from vizier._src.algorithms.designers.eagle_strategy import eagle_strategy
from vizier._src.algorithms.designers.eagle_strategy import eagle_strategy_utils
from vizier.benchmarks import experimenters

EagleStrategyDesiger = eagle_strategy.EagleStrategyDesigner
FireflyAlgorithmConfig = eagle_strategy.FireflyAlgorithmConfig
EagleStrategyUtils = eagle_strategy_utils.EagleStrategyUtils
FireflyPool = eagle_strategy_utils.FireflyPool
Firefly = eagle_strategy_utils.Firefly


def create_fake_trial(
    parent_fly_id: int,
    x_value: float,
    obj_value: Optional[float],
) -> vz.Trial:
  """Create a fake completed trial ('obj_value' = None means infeasible trial)."""
  trial = vz.Trial()
  measurement = vz.Measurement(
      metrics={
          eagle_strategy_utils.OBJECTIVE_NAME: vz.Metric(
              value=obj_value or float('inf')
          )
      }
  )
  trial.parameters['x'] = x_value
  trial.complete(
      measurement,
      inplace=True,
      infeasibility_reason='infeasible' if obj_value is None else None,
  )
  trial.metadata.ns('eagle')['parent_fly_id'] = str(parent_fly_id)
  return trial


def create_fake_problem_statement() -> vz.ProblemStatement:
  """Create a fake problem statement."""
  problem = vz.ProblemStatement()
  problem.search_space.root.add_float_param('x', 0.0, 10.0)
  problem.metric_information.append(
      vz.MetricInformation(
          name=eagle_strategy_utils.OBJECTIVE_NAME,
          goal=vz.ObjectiveMetricGoal.MAXIMIZE,
      )
  )
  return problem


def create_fake_fly(
    parent_fly_id: int,
    x_value: float,
    obj_value: Optional[float],
) -> Firefly:
  """Create a fake firefly with a fake completed trial."""
  trial = create_fake_trial(parent_fly_id, x_value, obj_value)
  return Firefly(id_=parent_fly_id, perturbation=1.0, generation=1, trial=trial)


def create_fake_empty_firefly_pool(capacity: int = 10) -> FireflyPool:
  """Create a fake empty Firefly pool."""
  problem = create_fake_problem_statement()
  # By default incorporating infeasible trials is disabled; setting it manually.
  config = FireflyAlgorithmConfig(infeasible_force_factor=0.1)
  rng = np.random.default_rng(0)
  utils = EagleStrategyUtils(problem_statement=problem, config=config, rng=rng)
  return FireflyPool(utils, capacity)


def create_fake_populated_firefly_pool(
    *,
    capacity: int,
    x_values: Optional[list[float]] = None,
    obj_values: Optional[list[Optional[float]]] = None,
    parent_fly_ids: Optional[list[int]] = None,
) -> FireflyPool:
  """Create a fake populated Firefly pool with a given capacity."""
  firefly_pool = create_fake_empty_firefly_pool(capacity=capacity)
  rng = np.random.default_rng(0)
  if not x_values:
    x_values = [float(x) for x in rng.uniform(low=0, high=10, size=(5,))]
  if not obj_values:
    obj_values = [
        float(o) for o in rng.uniform(low=-1.5, high=1.5, size=(len(x_values),))
    ]
  if not parent_fly_ids:
    parent_fly_ids = list(range(len(obj_values)))

  if not len(obj_values) == len(x_values) == len(parent_fly_ids):
    raise ValueError('Length of obj_values, ')

  for parent_fly_id, x_value, obj_value in zip(
      parent_fly_ids, x_values, obj_values
  ):
    firefly = create_fake_fly(
        parent_fly_id=parent_fly_id, x_value=x_value, obj_value=obj_value
    )
    # pylint: disable=protected-access
    firefly_pool._pool[parent_fly_id] = firefly

  # pylint: disable=protected-access
  firefly_pool._max_fly_id = capacity
  return firefly_pool


def create_fake_empty_eagle_designer() -> EagleStrategyDesiger:
  """Create a fake empty eagle designer."""
  problem = create_fake_problem_statement()
  return EagleStrategyDesiger(problem_statement=problem)


def create_fake_populated_eagle_designer(
    *,
    x_values: Optional[list[float]] = None,
    obj_values: Optional[list[Optional[float]]] = None,
    parent_fly_ids: Optional[list[int]] = None,
) -> EagleStrategyDesiger:
  """Create a fake populated eagle designer."""
  problem = create_fake_problem_statement()
  eagle_designer = EagleStrategyDesiger(problem_statement=problem)
  # pylint: disable=protected-access
  pool_capacity = eagle_designer._firefly_pool._capacity
  # Override the eagle designer's firefly pool with a populated firefly pool.
  eagle_designer._firefly_pool = create_fake_populated_firefly_pool(
      capacity=pool_capacity,
      x_values=x_values,
      obj_values=obj_values,
      parent_fly_ids=parent_fly_ids,
  )
  return eagle_designer


def create_continuous_exptr(func, dim=6):
  problem = experimenters.bbob.DefaultBBOBProblemStatement(dim)
  rng = np.random.default_rng(0)
  shift = rng.uniform(low=-2.0, high=2.0, size=(dim,))
  return experimenters.ShiftingExperimenter(
      exptr=experimenters.NumpyExperimenter(func, problem), shift=shift
  )


def create_continuous_log_scale_exptr(func, dim=6):
  problem = experimenters.bbob.DefaultBBOBProblemStatement(
      dim,
      scale_type=vz.ScaleType.LOG,
      min_value=1e1,
      max_value=1e4,
  )
  rng = np.random.default_rng(0)
  shift = rng.uniform(low=-1e2, high=0.0, size=(dim,))
  return experimenters.ShiftingExperimenter(
      exptr=experimenters.NumpyExperimenter(func, problem), shift=shift
  )


def create_categorical_exptr(num_params: int = 9, num_feasible_values: int = 5):
  return experimenters.L1CategorialExperimenter(
      num_categories=[num_feasible_values] * num_params, verbose=True
  )



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import numpy as np

from vizier._src.pyvizier.multimetric import hypervolume
from absl.testing import absltest


class HypervolumeTest(absltest.TestCase):

  def testParetoHypervolume(self):
    x = np.random.normal()
    y = np.random.normal()
    points = np.array([[x, y], [y, x]])
    origin = np.array([-1, -1])
    pf = hypervolume.ParetoFrontier(points, origin)
    self.assertAlmostEqual(
        pf.hypervolume(),
        max(2 * (x + 1) * (y + 1) - min(x + 1, y + 1)**2, 0.0),
        delta=0.3)

  def testParetoHypervolumeCumulative(self):
    points = np.random.normal(size=(100, 4))
    origin = np.zeros(4)

    pf = hypervolume.ParetoFrontier(points, origin)
    cumulative_vol = pf.hypervolume(is_cumulative=True)
    self.assertLen(cumulative_vol, len(points))
    for i in range(len(points)):
      if i < len(points) - 1:
        self.assertGreaterEqual(cumulative_vol[i + 1], cumulative_vol[i])
      else:
        self.assertAlmostEqual(cumulative_vol[i], pf.hypervolume())

  def testDimensionMismatchInit(self):
    points = np.array([[2, 1], [1, 2], [0, 0]])
    origin_3d = np.array([-1, -1, 1])
    # Dimension mismatch.
    with self.assertRaisesRegex(ValueError, 'Dimension mismatch'):
      hypervolume.ParetoFrontier(points, origin_3d)

  def testDimensionMismatchHypervolume(self):
    points = np.array([[2, 1], [1, 2], [0, 0]])
    origin_3d = np.array([-1, -1, 1])
    pf = hypervolume.ParetoFrontier(points, origin=np.array([0, 0]))
    # Dimension mismatch.
    with self.assertRaisesRegex(ValueError, 'Dimension mismatch'):
      pf.hypervolume(additional_points=origin_3d[..., np.newaxis])


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Output warper."""

import abc
import copy
from typing import Callable, Optional, Sequence, Tuple

import attr
import attrs
import equinox
import jax
from jax import numpy as jnp
import numpy as np
from scipy import stats
from tensorflow_probability.substrates import jax as tfp
from vizier._src.jax import types


tfb = tfp.bijectors


def _validate_labels(labels_arr: types.Array) -> types.Array:
  """Checks and modifies the shape and values of the labels."""
  labels_arr = labels_arr.astype(float)
  if not (labels_arr.ndim == 2 and labels_arr.shape[-1] == 1):
    raise ValueError(
        'Labels need to be an array of shape (num_points, 1).'
        f'Got shape: {labels_arr.shape}'
    )
  if np.isposinf(labels_arr).any():
    raise ValueError('Infinity metric value is not valid.')
  if np.isneginf(labels_arr).any():
    labels_arr[np.isneginf(labels_arr)] = np.nan
  return labels_arr


class OutputWarper(abc.ABC):
  """Interface for different output warper methods."""

  # TODO: implement the unwarp for different warper classes. Currently,
  # they return an error.

  @abc.abstractmethod
  def warp(self, labels_arr: types.Array) -> types.Array:
    """Runs the output warper of choice on an array of labels.

    Args:
      labels_arr: (num_points, 1) shaped array of unwarped labels. Note that
        each call accomodates one metric, which can be an objective or a safety
        metric. NaN and infinity values are allowed and can be warped based on
        the choice of warper. Labels are assumed to be maximizing. labels_arr
        will not be mutated.

    Returns:
      (num_points, 1) shaped array of finite warped mutated labels.
    """
    pass

  @abc.abstractmethod
  def unwarp(self, labels_arr: types.Array) -> types.Array:
    """Runs the inverse of output warper of choice on an array of labels.

    Args:
      labels_arr: (num_points, 1) shaped array of warped labels. Note tha each
        call accomodates one metric, which can be an objective or a safety
        metric. NaN and infinity values are allowed and will untouched. Labels
        are assumed to be maximizing. labels_arr will not be mutated.
        Technically any array can be passed to the unwarp method, but we are
        interested in unwarping warped labels for purposes of GP training and
        prediction.

    Returns:
      (num_points, 1) shaped array of finite unwarped mutated labels.
    """
    pass

  def warp_padded(self, labels: types.PaddedArray) -> types.PaddedArray:
    warped_unpadded_list = []
    unpadded = np.array(labels.unpad())
    for metric_id in range(labels.shape[-1]):
      warped_unpadded_list.append(self.warp(unpadded[:, metric_id]))

    warped_unpadded = np.concatenate(warped_unpadded_list, axis=-1)
    return labels.replace_array(warped_unpadded)


@attr.define
class BijectorWarper(OutputWarper):
  _bijector_factory: Callable[[types.Array], tfb.Bijector] = attr.ib
  _bijector: Optional[tfb.Bijector] = attr.ib(init=False, default=None)

  def warp(self, labels_arr: types.Array) -> types.Array:
    self._bijector = self._bijector_factory(labels_arr)
    return self._bijector(labels_arr)

  def unwarp(self, labels_arr: types.Array) -> types.Array:
    if self._bijector is None:
      raise ValueError('warp must be called first.')
    return self._bijector.inverse(labels_arr)


@attr.define
class OutputWarperPipeline(OutputWarper):
  """Performs a sequence of warpings on an input labels array."""

  warpers: Sequence[OutputWarper] = attr.ib(factory=list)

  def warp(self, labels_arr: types.Array) -> types.Array:
    """Sequntial warping of the labels.

    Note that if labels include one unique finite value, pipeline returns
    an array of zeros. Alternatively, if the labels include only infeasible
    values (eg. NaNs and -inf), pipeline returns an array of minus ones.
    Otherwise, it *sequntially* performs the warpings on labels. Warping is not
    done in place.

    Args:
      labels_arr: (num_points, 1) shaped array of labels.

    Returns:
      (num_points, 1) shaped array of warped labels.
    """
    labels_arr = copy.deepcopy(labels_arr)
    labels_arr = _validate_labels(labels_arr)
    if (
        np.isfinite(labels_arr).all()
        and len(np.unique(labels_arr).flatten()) == 1
    ):
      return np.zeros(labels_arr.shape)
    if np.isnan(labels_arr).all():
      return -1 * np.ones(shape=labels_arr.shape)
    for warper in self.warpers:
      labels_arr = warper.warp(labels_arr)
    return labels_arr

  def unwarp(self, labels_arr: types.Array) -> types.Array:
    """Sequential unwarping of the warped labels.

    Note that if labels include one unique finite value zero, the pipeline
    returns the array itself. Alternatively, if the labels include only minus
    ones, the pipeline returns an array of NaNs. Otherwise, it *sequentially*
    performs the unwarpings on labels. Unwarping is not done in place. This
    method accomodates inputs with duplicate entries and outputs equal output
    values for such entries.

    Args:
      labels_arr: (num_points, 1) shaped array of warped labels.

    Returns:
      (num_points, 1) shaped array of unwarped labels.
    """
    labels_arr = copy.deepcopy(labels_arr)
    labels_arr = _validate_labels(labels_arr)
    if (
        np.isfinite(labels_arr).all()
        and len(np.unique(labels_arr).flatten()) == 1
    ):
      if np.unique(labels_arr).item() == 0.0:
        return labels_arr
      if np.unique(labels_arr).item() == -1.0:
        return np.nan * np.ones(shape=labels_arr.shape)

    # reverse the order of warpers for unwarping to maintain bijective property.
    warpers = self.warpers[::-1]
    for warper in warpers:
      labels_arr = warper.unwarp(labels_arr)
    return labels_arr


def create_default_warper(
    *,
    half_rank_warp: bool = True,
    log_warp: bool = True,
    infeasible_warp: bool = True,
) -> OutputWarperPipeline:
  """Creates an output warper pipeline.

  Args:
    half_rank_warp: boolean indicating if half-rank warping to be performed.
    log_warp: boolean indicating if log warping to be performed.
    infeasible_warp: boolean indicating if infeasible warping to be performed.

  Returns:
    an instance of OutputWarperPipeline.
  """
  if not half_rank_warp and not log_warp and not infeasible_warp:
    raise ValueError(
        'At least one of "half_rank_warp", "log_warp" or "infeasible_warp" '
        'must be True.'
    )
  warpers = []
  if half_rank_warp:
    warpers.append(HalfRankComponent())
  if log_warp:
    warpers.append(LogWarperComponent())
  if infeasible_warp:
    warpers.append(InfeasibleWarperComponent())
  return OutputWarperPipeline(warpers)


def create_warp_outliers_warper(
    *,
    warp_outliers: bool = True,
    infeasible_warp: bool = True,
    transform_gaussian: bool = True,
) -> OutputWarperPipeline:
  """Creates an output warper outline which detects outliers and warps them."""
  warpers = []
  if warp_outliers:
    warpers.append(DetectOutliers())
  if infeasible_warp:
    warpers.append(InfeasibleWarperComponent())
  if transform_gaussian:
    warpers.append(TransformToGaussian())
  return OutputWarperPipeline(warpers)


@attr.define
class _HalfRankUnwarper:
  """Inverse mapping for HalfRankComponent.

  Attributes:
    _original_labels: Array of shape [N,]. Must be finite, unique, and sorted.
    _warped_labels: Array of shape [N,]. _warped_labels[i] is the result of
      applying the halfrank warper on `_original_labels[i]`.
  """

  # Labels must be unique and sorted.
  _original_labels: np.ndarray
  _warped_labels: np.ndarray
  _original_label_median: float

  def unwarp(self, label: float) -> float:
    """Unwarps the label."""
    # Anything above median is a no-op.
    if label >= self._original_label_median:
      return label

    # Try looking up the value
    # Use searchsorted and pull out three numbers.
    idx = np.searchsorted(self._warped_labels, label)
    candidates = self._warped_labels[
        max(0, idx - 1) : min(len(self._warped_labels), idx + 1)
    ]
    best_idx = np.argmin(np.abs(candidates - label))
    if np.isclose(self._warped_labels[best_idx], label):
      return self._original_labels[best_idx]

    # Label is smaller than the image of the warper.
    # Extrapolate.
    if label < np.min(self._warped_labels):
      return self._original_labels[0] - (
          np.abs(label - self._warped_labels[0])
          / (self._warped_labels[-1] - self._warped_labels[0])
      ) * (self._original_labels[-1] - self._original_labels[0])

    # All other cases: reverse the forward mapping.
    # Largest number less than `label`.
    # (which must exist because label >= min(warped_labels).)
    lower = np.searchsorted(self._warped_labels, label) - 1
    # Smallest number greater than `label`.
    # (which must exist because label < median.)
    upper = np.searchsorted(self._warped_labels, label)

    original_gap = self._original_labels[upper] - self._original_labels[lower]
    warped_gap = self._warped_labels[upper] - self._warped_labels[lower]
    return (
        self._original_labels[lower]
        + (label - self._warped_labels[lower]) * original_gap / warped_gap
    )


@attr.define
class HalfRankComponent(OutputWarper):
  """Warps half of an array of labels to fit into a Gaussian distribution.

  Note that this warping is performed on finite values of the array and NaNs are
  untouched.
  """

  _unwarper: Optional[_HalfRankUnwarper] = attr.field(default=None)

  def _estimate_std_of_good_half(
      self, unique_labels: np.ndarray, threshold: float
  ) -> float:
    """Estimates the standard devation of the good half of the array.

    Args:
      unique_labels: (num_points, 1) shaped array of unique labels.
      threshold: minimum label value to be considered in "good half".

    Returns:
      float estimated standard devation of the good half of the array.
    """
    good_half = unique_labels[unique_labels >= threshold]
    std = np.sqrt(
        ((good_half - threshold) ** 2).sum() * (1 / good_half.shape[0])
    )
    if std > 0:
      return std
    std = np.sqrt(
        ((unique_labels - threshold) ** 2).sum() * (1 / unique_labels.shape[0])
    )
    if np.isfinite(std):
      return std
    std = (np.abs(unique_labels - threshold)).sum() * (
        1 / unique_labels.shape[0]
    )
    return std

  def warp(self, labels_arr: types.Array) -> types.Array:
    """See base class."""
    labels_arr = _validate_labels(labels_arr)
    if labels_arr.size == 1:
      return labels_arr
    labels_arr = labels_arr.flatten()
    # Compute median, unique labels, and ranks.
    median = np.nanmedian(labels_arr)
    is_finite = np.isfinite(labels_arr)
    # np.unique also sorts the array.
    unique_labels, unique_idx = np.unique(
        labels_arr[is_finite], return_index=True
    )

    # Rank sort.
    ranks = stats.rankdata(labels_arr, method='dense')  # nans ranked last.
    dedup_median_index = unique_labels.searchsorted(median, 'left')
    denominator = (
        dedup_median_index + (unique_labels[dedup_median_index] == median) * 0.5
    )
    estimated_std = self._estimate_std_of_good_half(unique_labels, median)
    # Apply transformation to points below median.
    for i, (yy, rank) in enumerate(zip(labels_arr, ranks)):
      if np.isfinite(yy) and yy < median:
        # FYI: a lot of effort went into choosing this arbitrary
        # denominator. rankdata(method='max') and simply using
        # rank_quantile = rank / np.isfinite(labels_arr).sum() should be
        # just as fine.
        rank_quantile = 0.5 * (rank - 0.5) / denominator
        # rank_ppf is always less than 0
        rank_ppf = stats.norm.ppf(rank_quantile)
        labels_arr[i] = rank_ppf * estimated_std + median

    # Save information needed for unwarping.
    self._unwarper = _HalfRankUnwarper(
        original_labels=unique_labels,
        warped_labels=labels_arr[is_finite][unique_idx],
        original_label_median=unique_labels[len(unique_labels) // 2],
    )
    return labels_arr[:, np.newaxis]

  def unwarp(self, labels_arr: types.Array) -> types.Array:
    if self._unwarper is None:
      raise ValueError(' warp() needs to be called before unwarp() is called.')
    labels_arr = _validate_labels(labels_arr)
    labels_arr = labels_arr.flatten()
    if np.isnan(labels_arr).any():
      raise ValueError('unwarp does not support nan values.')

    for i in range(len(labels_arr)):
      labels_arr[i] = self._unwarper.unwarp(labels_arr[i])
    return labels_arr[:, np.newaxis]


@attr.define
class LogWarperComponent(OutputWarper):
  """Warps an array of labels to highlght the difference between good values.

  Note that this warping is performed on finite values of the array and NaNs are
  untouched.
  """

  _labels_min: Optional[float] = attr.field(default=None)
  _labels_max: Optional[float] = attr.field(default=None)
  offset: float = attr.field(default=1.5, validator=attrs.validators.gt(0.0))

  def warp(self, labels_arr: types.Array) -> types.Array:
    """See base class."""
    labels_arr = _validate_labels(labels_arr)
    self._labels_min = np.nanmin(labels_arr)
    self._labels_max = np.nanmax(labels_arr)
    labels_arr = labels_arr.flatten()
    finite_mask = np.isfinite(labels_arr)

    norm_diff = (self._labels_max - labels_arr[finite_mask]) / (
        self._labels_max - self._labels_min
    )
    labels_arr[finite_mask] = 0.5 - (
        np.log1p(norm_diff * (self.offset - 1)) / np.log(self.offset)
    )
    return labels_arr[:, np.newaxis]

  def unwarp(self, labels_arr: types.Array) -> types.Array:
    if self._labels_max is None or self._labels_min is None:
      raise ValueError(' warp() needs to be called before unwarp() is called.')
    labels_arr = labels_arr.flatten()
    labels_arr = self._labels_max - (
        np.exp(np.log(self.offset) * (0.5 - labels_arr)) - 1
    ) * (self._labels_max - self._labels_min) / (self.offset - 1)
    return labels_arr[:, np.newaxis]


@attr.define
class InfeasibleWarperComponent(OutputWarper):
  """Warps the infeasible/nan value to feasible/finite values.

  This OutputWarper handles a mix of feasible and infeasible (NaN) labels, and
  returns feasible labels.

  Typically, it's the last element in an `OutputWarperPipeline`.
  What it does:
    * warp returns a feasible value.
    * If we have a mix of feasible values and infeasible values, warp() applied
      to any of the feasible values will be larger than warp() applied to an
      infeasible value.
    * The expected value of warp() over the set of feasible values will be zero.

  Recall that we typically use this class with a Gaussian Process regressor
  which has a prior mean of zero, and that the GP regressor's prediction
  approaches zero when it's far away from any support points. So, a zero output
  should correspond to the overall average of all label values.

  In other words, the expected value of warp() should be zero, when feasible and
  infeasible values are chosen with frequencies estimated from the set of all
  available values. I.e.
  0 = p_feasible * Expected(v_feasible) + (1 - p_feasible) * v_unfeasible,
  where p_feasible is the probability of getting a feasible value,
  Expected(v_feasible) is the expected value of warp() over feasible values,
  and v_unfeasible is the value we assign to infeasible values.

  This means that when we have mostly feasible values, we want
  Expected(v_feasible) near zero; and when we have mostly infeasible values,
  we want Expected(v_feasible) to be positive.

  This is entirely reasonable: when feasible values are rare, they are
  individually more surprising and should affect the regressor's model
  more, and vice versa.  Or, equivalently, when feasible values are very
  rare, and when the GP regressor is evaluated far away from any support
  points, the GP regressor will predict a value (i.e. zero) that should be
  nearly equal to v_infeasible.
  """

  _shift: Optional[float] = attr.field(default=None)

  def warp(self, labels_arr: types.Array) -> types.Array:
    labels_arr = _validate_labels(labels_arr)
    labels_arr = labels_arr.flatten()

    if np.isnan(labels_arr).all():
      # Edge case when all values are NaN.
      self._shift = np.nan
      labels_arr[:] = 0
      return labels_arr[:, np.newaxis]

    labels_range = np.nanmax(labels_arr) - np.nanmin(labels_arr)
    warped_bad_value = np.nanmin(labels_arr) - (0.5 * labels_range + 1)
    num_feasible = labels_arr.size - np.isnan(labels_arr).sum()
    # With the data we have available, we can estimate the relative frequency
    # of feasible (p_feasible) and unfeasible points (p_unfeasible).
    # We use the Jeffrey's version of Laplace Smoothing: see
    # https://en.wikipedia.org/wiki/Jeffreys_prior#N-sided_die_with_biased_probabilities
    # and
    # https://en.wikipedia.org/wiki/Additive_smoothing. The basic logic being
    # that even though we may have no observations of (e.g.) a feasible point,
    # we believe that feasible points are possible, and so we are reluctant to
    # assign them a zero probability.
    p_feasible = (0.5 + num_feasible) / (1 + labels_arr.size)
    self._shift = -np.nanmean(labels_arr) * p_feasible - warped_bad_value * (
        1 - p_feasible
    )
    labels_arr[np.isnan(labels_arr)] = warped_bad_value
    labels_arr[~np.isnan(labels_arr)] += self._shift
    return labels_arr[:, np.newaxis]

  def unwarp(self, labels_arr: types.Array) -> types.Array:
    if self._shift is None:
      raise ValueError('warp() needs to be called before unwarp() is called.')
    return labels_arr - self._shift


class ZScoreLabels(OutputWarper):
  """Sandardizes finite label values, leaving the NaNs & infinities out."""

  def warp(self, labels_arr: types.Array) -> types.Array:
    """Sandardizes finite label values to scale the mean to 0 and std to 1.

    Args:
      labels_arr: (num_points, 1) shaped array of labels.

    Returns:
      (num_points, 1) shaped array of standardize labels.
    """
    labels_arr = _validate_labels(labels_arr)
    if np.isnan(labels_arr).all():
      raise ValueError('Labels need to have at least one non-NaN entry.')
    labels_finite_ind = np.isfinite(labels_arr)
    labels_arr_finite = labels_arr[labels_finite_ind]
    if np.nanstd(labels_arr_finite) == 0 or not np.isfinite(
        np.nanstd(labels_arr_finite)
    ):
      return labels_arr
    labels_arr_finite_normalized = (
        labels_arr_finite - np.nanmean(labels_arr_finite)
    ) / np.nanstd(labels_arr_finite)
    labels_arr[labels_finite_ind] = labels_arr_finite_normalized
    return labels_arr

  def unwarp(self, labels_arr: types.Array) -> types.Array:
    raise NotImplementedError(
        'unwarp  method for ZScoreLabels is not implemented yet.'
    )


@attr.define
class NormalizeLabels(OutputWarper):
  """Normalizes the finite label values, leaving the NaNs & infinities out."""

  target_interval: Tuple[float, float] = attr.ib(default=(0.0, 1.0))

  def __attrs_post_init__(self):
    if self.target_interval[0] > self.target_interval[1]:
      raise ValueError(f'Bounds {self.target_interval} is invalid.')

  def warp(self, labels_arr: types.Array) -> types.Array:
    """Normalizes the finite label values to bring them within target_interval.

    If all finite labels are equal, they get mapped to target_interval's
    midpoint.

    Args:
      labels_arr: (num_points, 1) shaped array of labels.

    Returns:
      (num_points, 1) shaped array of normalized labels.
    """
    labels_arr = _validate_labels(labels_arr)
    if np.isnan(labels_arr).all():
      raise ValueError('Labels need to have at least one non-NaN entry.')

    labels_finite_ind = np.isfinite(labels_arr)
    labels_arr_finite = labels_arr[labels_finite_ind]

    source_interval = (np.min(labels_arr_finite), np.max(labels_arr_finite))

    if source_interval[0] == source_interval[1]:
      midpoint = (self.target_interval[0] + self.target_interval[1]) / 2
      labels_arr[labels_finite_ind] = midpoint
      return labels_arr

    labels_arr_finite_normalized = np.interp(
        labels_arr_finite, source_interval, self.target_interval
    )
    labels_arr[labels_finite_ind] = labels_arr_finite_normalized
    return labels_arr

  def unwarp(self, labels_arr: types.Array) -> types.Array:
    raise NotImplementedError(
        'unwarp  method for NormalizeLabels is not implemented yet.'
    )


@attr.define
class DetectOutliers(OutputWarper):
  """Detects outliers from an array of labels.

  The goal of this warper is to detect the unreasonably bad labels, aka
  outliers. Eg, assuming a maximization problem where the `normal` range of
  labels is [1, 10], the occasionally observed value of -10**76 counts as an
  outlier. For this warper, we only consider the *finite* values in the labels
  and leave the NaN and infinity values untouched.

  The proposed warping finds the difference between the maximum label and the
  median of the labels.

  Attributes:
    min_zscore: number of stds below the median for variance estimation.
    max_zscore: number of stds above the median for outlier detection.
  """

  min_zscore: float = attr.field(kw_only=True, default=6.0)
  max_zscore: float = attr.field(kw_only=True, default=None)

  def _estimate_variance(self, labels_arr: types.Array) -> float:
    """Estimates the variance of labels array using the top half values.

    The estimation is a function of the size of the labels array. For details
    and derivations, see the following paper:
    `Estimating the mean and variance from the median, range, and the size of
    a sample` paper found below
    https://link.springer.com/content/pdf/10.1186/1471-2288-5-13.pdf.

    Args:
      labels_arr: (num_points, 1) shaped array of labels.

    Returns:
      The estiamated variance of the labels.
    """
    num_points = labels_arr.shape[0]
    labels_median = np.nanmedian(labels_arr)
    labels_max = np.nanmax(labels_arr)

    if not np.isfinite(labels_max) or np.isnan(labels_max):
      raise ValueError('The max label value should be finite.')

    if not np.isfinite(labels_median) or np.isnan(labels_median):
      raise ValueError('The median label value should be finite.')

    if self.max_zscore:
      return (labels_max - labels_median) / self.min_zscore

    if num_points >= 70:
      return (labels_max - labels_median) / 3
    elif num_points >= 15:
      return (labels_max - labels_median) / 2
    else:
      # We hallucinate the min labels value
      labels_min_hallucinated = labels_median - np.max(labels_arr)
      # Following the paper assumptions, we need to make sure the hallucinated
      # min is non-negative and shift the rest of the values accordingly.
      if labels_min_hallucinated < 0:
        labels_min_hallucinated = 0
        labels_max -= labels_min_hallucinated
        labels_median -= labels_min_hallucinated

      # Eq. 12 in the paper mentioned above.
      a, m, b = labels_min_hallucinated, labels_median, labels_max
      n = num_points

      out = a**2 + m**2 + b**2
      out += ((n - 3) / 2) * ((a + m) ** 2 + (b + m) ** 2) / 4
      out -= n * ((a + 2 * m + b) / 4 + (a - 2 * m + b) / (4 * n)) ** 2
      return out / (n - 1)

  def warp(self, labels_arr: types.Array) -> types.Array:
    labels_arr = _validate_labels(labels_arr)
    labels_finite_ind = np.isfinite(labels_arr)
    labels_arr_finite = labels_arr[labels_finite_ind]
    labels_median = np.median(labels_arr_finite)
    labels_std = np.sqrt(self._estimate_variance(labels_arr_finite))
    threshold = labels_median - self.min_zscore * labels_std
    labels_arr_finite[labels_arr_finite < threshold] = np.nan
    labels_arr[labels_finite_ind] = labels_arr_finite
    return labels_arr

  def unwarp(self, labels_arr: types.Array) -> types.Array:
    raise NotImplementedError(
        'unwarp  method for DetectOutliers is not implemented yet.'
    )


class TransformToGaussian(OutputWarper):
  """Transforms the labels into a Gaussian distribution.

  The goal of this warper is to transform the label into a Gaussian sample to
  better suit it for a Gaussian process. Here, we use a non-parametric warper
  called quantile transformer. Traditonally, this transformer relies on the
  relational rank of the labels, however this approach does not take into
  account the values of the labels. To address this issue, we also provide
  another option to use the normalized distances instead.

  Alternatives include a box-cox transformation (for non-negative labels) or a
  yeo-johnson transformation. Both these methods are parameteric and their
  parameters can be learned using a maximum likelihood approach.
  """

  def __init__(
      self,
      *,
      softclip_low: float = 1e-10,
      softclip_high: float = 1 - 1e-10,
      softclip_hinge_softness: float = 0.01,
      use_rank: bool = False,
  ):
    self.softclip_low = softclip_low
    self.softclip_high = softclip_high
    self.softclip_hinge_softness = softclip_hinge_softness
    self.use_rank = use_rank

  def warp(self, labels_arr: types.Array) -> types.Array:
    labels_arr = _validate_labels(labels_arr)
    labels_arr = np.asarray(labels_arr, dtype=np.float64)
    labels_arr_flattened = labels_arr.flatten()
    if self.use_rank:
      base_for_transform = np.argsort(labels_arr_flattened)
    else:
      base_for_transform = labels_arr_flattened
    base_for_transform_normalized = (
        base_for_transform - np.min(base_for_transform)
    ) / (np.max(base_for_transform) - np.min(base_for_transform))
    clip = tfp.bijectors.SoftClip(
        low=np.array(self.softclip_low, dtype=labels_arr.dtype),
        high=np.array(self.softclip_high, dtype=labels_arr.dtype),
        hinge_softness=self.softclip_hinge_softness,
    )
    base_for_transform_normalized_clipped = np.array(
        clip.forward(base_for_transform_normalized)
    )
    normal_dist = tfp.distributions.Normal(0.0, 1)
    labels_arr_transformed = normal_dist.quantile(
        base_for_transform_normalized_clipped
    )
    labels_arr_transformed = np.reshape(
        labels_arr_transformed, labels_arr.shape
    )
    return labels_arr_transformed

  def unwarp(self, labels_arr: types.Array) -> types.Array:
    raise NotImplementedError(
        'unwarp  method for TransformToGaussian is not implemented yet.'
    )


class LinearOutputWarper(equinox.Module):
  """Linear output warper.

  The LinearOutputWarper applies affine transformation to transform the labels
  to [low_bound, high_bound].

  Note: low_bound, high_bound are scalars.
  """

  low_bound: types.Array  # shape: ()
  high_bound: types.Array  # shape: ()
  min_value: types.Array  # shape: (num_metrics,)
  max_value: types.Array  # shape: (num_metrics,)

  @classmethod
  def from_obs(
      cls, y_obs: types.Array, low_bound: float = -2.0, high_bound: float = 2.0
  ) -> 'LinearOutputWarper':
    # y shape: (num_samples, num_metrics)
    min_value = jnp.min(y_obs, axis=0)
    max_value = jnp.max(y_obs, axis=0)

    return cls(
        low_bound=jnp.asarray(low_bound, dtype=jnp.float64),
        high_bound=jnp.asarray(high_bound, dtype=jnp.float64),
        min_value=min_value,
        max_value=max_value,
    )

  @property
  def _bijector(self) -> tfb.Bijector:
    # The linear transformation is:
    # norm_y = (y - min_value) / (max_value - min_value)
    # y --> norm_y * (self.high_bound - self.low_bound) + self.low_bound
    return tfb.Chain([
        tfb.Shift(self.low_bound),
        self._slope_bijector,
        tfb.Shift(-self.min_value),
    ])

  @property
  def _slope_bijector(self) -> tfb.Bijector:
    return tfb.Scale(
        (self.high_bound - self.low_bound) / (self.max_value - self.min_value),
    )

  def warp(self, y: types.Array) -> jax.Array:
    """Warp the y values into [low_bound, high_bound]."""
    # y shape: (num_samples, num_metrics)
    return self._bijector.forward(y)

  def unwarp(self, y: types.Array) -> jax.Array:
    """Un-warp the y values into [min_value, max_value]."""
    # y shape: (num_samples, num_metrics)
    return self._bijector.inverse(y)

  def unwarp_stddev(self, warped_stddev: types.Array) -> jax.Array:
    """Un-warp the standard deviation."""
    return self._slope_bijector.inverse(warped_stddev)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for sql_datastore."""
import os
import sqlalchemy as sqla

from vizier._src.service import constants
from vizier._src.service import datastore_test_lib
from vizier._src.service import sql_datastore
from vizier._src.service.testing import util as test_util
from absl.testing import absltest


class SQLDataStoreTest(datastore_test_lib.DataStoreTestCase):

  def setUp(self):
    self.owner_id = 'my_username'
    self.study_id = '123123123'
    self.client_id = 'client_0'
    self.example_study = test_util.generate_study(self.owner_id, self.study_id)
    self.example_trials = test_util.generate_trials(
        [1, 2], owner_id=self.owner_id, study_id=self.study_id
    )
    self.example_suggestion_operations = (
        test_util.generate_suggestion_operations(
            [1, 2, 3, 4], self.owner_id, self.study_id, self.client_id
        )
    )
    self.example_early_stopping_operations = (
        test_util.generate_early_stopping_operations(
            [1, 2], self.owner_id, self.study_id
        )
    )

    engine = sqla.create_engine(constants.SQL_MEMORY_URL, echo=True)
    self.datastore = sql_datastore.SQLDataStore(engine)
    super().setUp()

  def test_study_api(self):
    self.assertStudyAPI(self.datastore, self.example_study)

  def test_trial(self):
    self.assertTrialAPI(self.datastore, self.example_study, self.example_trials)

  def test_suggestion_operation(self):
    self.assertSuggestOpAPI(
        self.datastore,
        self.example_study,
        self.client_id,
        self.example_suggestion_operations,
    )

  def test_early_stopping_operation(self):
    self.assertEarlyStoppingAPI(
        self.datastore,
        self.example_study,
        self.example_trials,
        self.example_early_stopping_operations,
    )

  def test_update_metadata(self):
    self.assertUpdateMetadataAPI(
        self.datastore, self.example_study, self.example_trials
    )


class SQLDataStoreAdditionalTest(absltest.TestCase):
  """For additional tests outside of regular database functionality."""

  def setUp(self):
    super().setUp()
    self.owner_id = 'my_username'
    self.study_id = '123123123'
    self.example_study = test_util.generate_study(self.owner_id, self.study_id)

  @absltest.skip("Github workflow tests don't allow using directories.")
  def test_local_hdd_persistence(self):
    db_path = os.path.join(absltest.get_default_test_tmpdir(), 'local.db')
    sql_url = f'sqlite:///{db_path}'

    engine = sqla.create_engine(sql_url, echo=True)
    datastore = sql_datastore.SQLDataStore(engine)
    datastore.create_study(self.example_study)
    del datastore

    engine2 = sqla.create_engine(sql_url, echo=True)
    datastore2 = sql_datastore.SQLDataStore(engine2)
    study = datastore2.load_study(self.example_study.name)

    self.assertEqual(self.example_study, study)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Random designer."""

from typing import Any, Optional, Sequence

import numpy as np
from vizier import algorithms as vza
from vizier import pyvizier as vz
from vizier.pyvizier import converters


class RandomDesigner(vza.Designer):
  """Random designer.

  This designer uniformly samples from the scaled search space.
  """

  def __init__(self,
               search_space: vz.SearchSpace,
               *,
               dtype: Any = np.float64,
               seed: Optional[int] = None):
    """Init.

    Args:
      search_space: Must be a flat search space.
      dtype:
      seed: Any valid seed for np.random.RandomState.
    """
    if search_space.is_conditional:
      # TODO: Add conditional sampling case.
      raise ValueError(
          f'This designer {self} does not support conditional search.')

    def create_input_converter(pc):
      # Setting 'max_discrete_indices' to not continuify DISCRETE parameters.
      return converters.DefaultModelInputConverter(
          pc, scale=True, max_discrete_indices=np.inf, float_dtype=dtype
      )

    self._converter = converters.DefaultTrialConverter(
        [create_input_converter(pc) for pc in search_space.parameters])

    self._rng = np.random.RandomState(seed)

  @classmethod
  def from_problem(
      cls, problem: vz.ProblemStatement, seed: Optional[int] = None
  ):
    """For wrapping via `PartiallySerializableDesignerPolicy`."""
    return RandomDesigner(problem.search_space, seed=seed)

  def update(
      self, completed: vza.CompletedTrials, all_active: vza.ActiveTrials
  ) -> None:
    pass

  def suggest(self,
              count: Optional[int] = None) -> Sequence[vz.TrialSuggestion]:
    """Make new suggestions.

    Args:
      count: Makes best effort to generate this many suggestions. If None,
        suggests as many as the algorithm wants.

    Returns:
      New suggestions.
    """
    count = count or 1
    sample = dict()
    for name, spec in self._converter.output_specs.items():
      if spec.type == converters.NumpyArraySpecType.DISCRETE:
        sample[name] = self._rng.random_integers(
            spec.bounds[0], spec.bounds[1] - spec.num_oovs, size=[count, 1])
      elif spec.type == converters.NumpyArraySpecType.CONTINUOUS:
        # Trial-Numpy converter was configured to scale values to [0, 1].
        # Sample uniformly in that range and convert back to Trials.
        sample[name] = self._rng.random([count, 1])
      else:
        raise ValueError(f'Unsupported type: {spec.type} in {spec}')
    return [
        vz.TrialSuggestion(p) for p in self._converter.to_parameters(sample)
    ]
